/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2008 Bull S.A.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 * --------------------------------------------------------------------------
 * $Id: ConfigurationProvider.java 4389 2008-12-15 13:48:57Z alitokmen $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.launcher;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.apache.felix.framework.cache.BundleCache;
import org.apache.felix.main.AutoActivator;

/**
 * Configuration provider to start shelbie in standalone mode.
 * @author Benjamin Strappazzon.
 */
public class ConfigurationProvider {

    /**
     * Felix configuration file. Define which system packages to start.
     */
    private static final String FELIX_CONFIG_FILE = "default-config.properties";

    /**
     * System property name to define shelbie root path.
     */
    private static final String MAVEN_ROOT = "repositories/bundles";

    /**
     * Default felix bundle cache directory.
     */
    private static final String FELIX_BUNDLE_CACHE = "felix/cache";

    /**
     * List of excluded bundles.
     */
    private List<String> excluded;

    /**
     * Root path of maven repository.
     */
    private String mavenRootPath;

    /**
     * Properties that describes which bundles must be started.
     */
    private Properties autoDeployConfigurationFile;

    /**
     * Felix configuration.
     */
    private Properties felixConfig;

    /**
     * Location of the felix cache.
     */
    private String felixCacheDirectory;

    /**
     * Construct a Configuration provider with default settings.
     * @throws IOException Thrown if felix configuration file cannot be read.
     */
    public ConfigurationProvider() throws IOException {
        this(null, null, null, null, null);
    }

    /**
     * Construct a launcher with a list of bundle to be automatically started.
     * @param autoDeployConfigurationFile List of bundles to be started.
     * @throws IOException Thrown if felix configuration file cannot be read.
     */
    public ConfigurationProvider(final Properties autoDeployConfigurationFile) throws IOException {
        this(autoDeployConfigurationFile, null, null, null, null);
    }

    /**
     * Construct a launcher with a list of bundle to be automatically started
     * except those present in the excluded list.
     * @param autoDeployConfigurationFile List of bundles to be started.
     * @param excluded A list of bundle to be excluded.
     * @throws IOException Thrown if felix configuration file cannot be read.
     */
    public ConfigurationProvider(final Properties autoDeployConfigurationFile, final List<String> excluded) throws IOException {
        this(autoDeployConfigurationFile, excluded, null, null, null);
    }

    /**
     * Construct a launcher with customized options.
     * @param autoDeployConfigurationFile List of bundles to be auto started.
     * @param excluded In the list, bundles that must not be started.
     * @param felixConfig A customized felix configuration.
     * @param mavenRootPath The path to the maven repository.
     * @param felixCache Felix bundle cache.
     * @throws IOException Thrown if felix configuration file cannot be read.
     */
    public ConfigurationProvider(final Properties autoDeployConfigurationFile, final List<String> excluded,
            final Properties felixConfig, final String mavenRootPath, final String felixCache) throws IOException {
        super();
        this.felixConfig = new Properties();

        if (autoDeployConfigurationFile == null) {
            // empty auto deploy list
            this.autoDeployConfigurationFile = new Properties();
        } else {
            this.autoDeployConfigurationFile = autoDeployConfigurationFile;
        }

        if (excluded == null) {
            // empty list
            this.excluded = new ArrayList<String>();
        } else {
            this.excluded = excluded;
        }

        if (felixConfig == null) {
            this.felixConfig.load(Launcher.class.getResourceAsStream(ConfigurationProvider.FELIX_CONFIG_FILE));
        } else {
            this.felixConfig = felixConfig;
        }

        if (mavenRootPath == null) {
            // use default settings to configure felix framework
            this.mavenRootPath = ConfigurationProvider.MAVEN_ROOT;
        } else {
            this.mavenRootPath = mavenRootPath;
        }

        if (felixCache == null) {
            //use default settings
            this.felixCacheDirectory = ConfigurationProvider.FELIX_BUNDLE_CACHE;
        } else {
            this.felixCacheDirectory = felixCache;
        }

    }

    /**
     * Compute a felix configuration with provided settings.
     * @return a Felix specific configuration.
     * @throws IOException configuration not found
     */
    @SuppressWarnings("unchecked")
    public Map getConfiguration() throws IOException {

        Properties prop = this.felixConfig;
        //add cache directory to the felix config
        prop.put(BundleCache.CACHE_PROFILE_DIR_PROP, this.felixCacheDirectory);
        findAutoDeployedBundles(prop);
        return prop;

    }

    /**
     * Update the given config with start level options.
     * @param config configuration to enhance.
     * @throws IOException if default property file is not found.
     */
    private void findAutoDeployedBundles(final Properties config) throws IOException {

        // Felix bundles configuration properties
        Properties levels = this.autoDeployConfigurationFile;

        // Iterates over the properties
        for (Iterator<Object> it = levels.keySet().iterator(); it.hasNext();) {

            // Process level properties
            String key = (String) it.next();

            if (key.startsWith("install.level.")) {
                // Get the associated level
                String level = key.substring("install.level".length());

                // Configure the bundle list to install
                config.put(AutoActivator.AUTO_INSTALL_PROP.concat(level), getBundleList(levels, key));
            }

            if (key.startsWith("start.level.")) {
                // Get the associated level
                String level = key.substring("start.level".length());

                // Configure the bundle list to start
                config.put(AutoActivator.AUTO_START_PROP.concat(level), getBundleList(levels, key));
            }
        }
    }

    /**
     * Get the bundle list for a given property.
     * @param levels The property list
     * @param key The searched key
     * @return The bundle list for a given property.
     * @throws IOException If the list cannot be built.
     */
    private String getBundleList(final Properties levels, final String key) throws IOException {
        String value = levels.getProperty(key);
        StringBuilder sb = new StringBuilder();

        // The bundle list separator is ','
        String[] bundles = value.split(",");
        for (int i = 0; i < bundles.length; i++) {
            String bundle = bundles[i];

            // Only process non-empty parts
            if (!"".equals(bundle)) {

                // The bundle is specified using the following format:
                // <groupId>:<artifactId>[:<version>]
                String[] artifact = bundle.split(":");

                String groupId = artifact[0].trim();
                String artifactId = artifact[1].trim();
                String version = null;
                String classifier = null;

                if (!this.excluded.contains(groupId + ":" + artifactId)) {
                    // This element is not excluded
                    if (artifact.length == 3) {
                        // Is the third element a version or a
                        // classifier ?
                        classifier = getClassifier(artifact[2]);
                        if (classifier == null) {
                            // this is NOT a classifier
                            version = artifact[2].trim();
                        }
                    } else if (artifact.length == 4) {
                        // We have both version + classifier
                        // first is the version
                        version = artifact[2].trim();
                        // then, the classifier
                        classifier = getClassifier(artifact[3]);
                        if (classifier == null) {
                            // in this case, this is an error
                            throw new IOException("Incorrect classifier in bundle: " + bundle);
                        }
                    } else if (artifact.length > 4) {
                        // More elements, invalid
                        throw new IOException("Incorrect number of parts in bundle: " + bundle);
                    }

                    // Use the reference attribute to load the bundle
                    // from the local file without copying it into the
                    // cache of Felix
                    sb.append("reference:");
                    String bundleReference = getBundleMaven2Location(mavenRootPath, groupId, artifactId, version, classifier)
                            .concat(" ");
                    if (bundleReference == null) {
                        throw new IllegalArgumentException("Unable to find a bundle reference for '" + groupId + "'/'"
                                + artifactId + "'/'" + version + "'/'" + classifier + "'.");
                    }
                    sb.append(bundleReference);
                } // Else this element is excluded, do nothing
            }
        }

        return sb.toString();
    }

    /**
     * @param value value from which the classifier will be extracted
     * @return the classifier value if this is a valid classifier, null
     *         otherwise
     */
    private String getClassifier(final String value) {
        String classifier = null;
        String trimmed = value.trim();
        if (trimmed.startsWith("{") && trimmed.endsWith("}")) {
            // classifier
            classifier = trimmed.substring(1, trimmed.length() - 1);
        }
        return classifier;
    }

    /**
     * Create the artifact jar name.
     * @param artifactId Id if the artifact.
     * @param version Artifact's version
     * @param classifier Artifact's classifier (may be null)
     * @return a correct jar filename.
     */
    private String getArtifactPath(final String artifactId, final String version, final String classifier) {

        String path = artifactId.concat("-").concat(version);
        if (classifier != null) {
            path = path.concat("-").concat(classifier);
        }

        path = path.concat(".jar");
        return path;
    }

    /**
     * @param repository Artifact repository
     * @param groupId Artifact groupId
     * @param artifactId Artifact ID
     * @param version Artifact version
     * @param classifier Artifact's classifier (may be null)
     * @return a string-ified URL
     * @throws MalformedURLException If jar file is not in the repository.
     */
    public static String getBundleMaven2Location(final String repository, final String groupId, final String artifactId,
            final String version, final String classifier) throws MalformedURLException {

        // Creates a File from artifact info
        String appendedClassifier = "";
        if (classifier != null) {
            // Handle classifier if any
            appendedClassifier = appendedClassifier.concat("-").concat(classifier);
        }
        String path = constructDirectoryPath(groupId, artifactId, version);
        File bundle = new File(repository, path.concat(File.separator).concat(artifactId).concat("-").concat(version).concat(
                appendedClassifier).concat(".jar"));

        // Transform to URI first, because File.toURL() is buggy
        return bundle.toURI().toURL().toExternalForm();
    }

    /**
     * @param groupId Artifact groupId
     * @param artifactId Artifact ID
     * @param version Artifact version
     * @return transform the artifact coordinates into a m2 repository path
     */
    private static String constructDirectoryPath(final String groupId, final String artifactId, final String version) {
        // Transform groupId '.' into '/'
        String transformed = groupId.replace('.', File.separatorChar);
        return transformed.concat(File.separator).concat(artifactId).concat(File.separator).concat(version);
    }
}
